import logging
import os
import sys
from pathlib import Path

from .android.adb import switch_to_tcpip
from .nimses import main

if __name__ == '__main__':
    """
    Launches nimses-adimate.
    This launch script appends the platform-tools directory to the end of the PATH variable in order for certain 
    operating systems to find the `adb` executable.
    Then it switches ADB to "tcpip" mode and launches the actual main code.
    """
    logging.basicConfig(stream=sys.stdout,
                        level=logging.INFO,
                        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
    os.environ["PATH"] += os.pathsep + str(Path(__file__).parent) + os.sep + "platform-tools"

    device = switch_to_tcpip()
    main(device)



import logging
import socket
from concurrent.futures import Executor
from time import time

import cv2
import numpy as np
from adbutils import AdbDevice

from .raw_socket import send_raw

logger = logging.getLogger(__name__)


def screencap_raw(device: AdbDevice):
    start = time()
    buffer = send_raw(device, "framebuffer:", 20 * 1024 * 1024)
    logger.debug("Recv RAW took %f seconds" % (time() - start))

    # Unpack the header
    version, bpp, colorspace, size, width, height, \
    red_offset, red_length, \
    blue_offset, blue_length, \
    green_offset, green_length, \
    alpha_offset, alpha_length = np.frombuffer(buffer, "uint32", 14)

    return cv2.cvtColor(np.reshape(np.frombuffer(buffer[14 * 4:], "uint8"), (height, width, 4)), cv2.COLOR_RGB2BGR)


def screencap_png(device: AdbDevice) -> np.ndarray:
    start = time()
    buffer = send_raw(device, "exec:screencap -p", 1024 * 1024)
    logger.debug("Recv PNG took %f seconds" % (time() - start))

    return cv2.imdecode(np.frombuffer(buffer, "uint8"), cv2.IMREAD_UNCHANGED)


def screencap_h264(device: AdbDevice, executor: Executor):
    start = time()
    device.push("../scrcpy-win64/scrcpy-server.jar", "/data/local/tmp/scrcpy-server.jar")
    print("Pushed server component")
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    port = device.adb_output("reverse", "tcp:5556", "localabstract:scrcpy")
    print("[ADB]", port)
    # sock.bind(("127.0.0.1", 5556))
    # sock.listen()

    executor.submit(lambda: print("[SCRCPY]", device.shell("CLASSPATH=/data/local/tmp/scrcpy-server.jar "
                                                           "app_process / com.genymobile.scrcpy.Server 0 8000000 "
                                                           "false - false true")))
    print("Executed scrcpy")

    print("Waiting for connection")
    # conn, info = sock.accept()
    # print(info)
    # print(conn.recv(1))
    # print(conn.recv(64))
    # print(conn.recv(2))
    # print(conn.recv(2))
    # print("Closing transmission")
    # device.shell("\x03")
    # conn.close()
    # sock.close()
    print("Recv h264 took %f seconds" % (time() - start))



import logging
from concurrent.futures import Future
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from time import time, sleep, strftime, localtime
from typing import Tuple

import cv2
import numpy as np
from adbutils import adb, AdbDevice

from .android.async_helper import log_error
from .android.screencap import screencap_png

# CONSTANTS
PLAYSTORE_PACKAGE = "com.android.vending"
NIMSES_PACKAGE = "com.nimses"
NIMSES_MAIN_ACTIVITY = "com.nimses.navigation.presentation.view.screens.main.MainActivity"
NIMSES_AD_ACTIVITIES = (
    "com.adcolony.sdk.AdColonyInterstitialActivity",
    "com.unity3d.ads.adunit.AdUnitActivity",
    "com.unity3d.services.ads.adunit.AdUnitActivity",
    "com.ironsource.sdk.controller.ControllerActivity"
)

TEMPLATE_AD_IMAGE = cv2.imread(str(Path(__file__).parent) + "/templates/ads_inline_fhd_de.png", cv2.IMREAD_GRAYSCALE)

logger = logging.getLogger("NAD")


def screen_shot_android(device: AdbDevice) -> np.ndarray:
    """
    Takes a screen shot of the given Android device.
    This method just exists to easily change the implementation of the screen shot mechanism.

    :param device: the device to take a screen shot from
    :return: the screen shot in a raw `ndarray`
    """
    return screencap_png(device)


def find_image(image: np.ndarray, search: np.ndarray, threshold=0.05, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
    """
    Finds an image inside another image with a maximum threshold and additional arguments.
    Uses `cv2.matchTemplate` with the `cv2.TN_SQDIFF_NORDME` method.

    :param image: the source image to search in; converted to grayscale for opencv.
    :param search: the image to search for in the source image.
    :param threshold: the maximum threshold for the source (default: 0.05)
    :param kwargs: additional arguments to `cv2.matchTemplate`
    :return: two `ndarray`s, one with the matches and one with the entire result image (grayscale, the darker the pixel,
    the more closely it match es the search image)
    """
    img_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    res = cv2.matchTemplate(img_gray, search, cv2.TM_SQDIFF_NORMED, **kwargs)
    return np.where(res <= threshold), res


def find_inline_ads(image: np.ndarray) -> Tuple[Tuple[int, int], np.ndarray]:
    """
    Finds and marks Nimses ads in a given image.
    Uses the `TEMPLATE_AD_IMAGE`.

    :param image: the image (screen shot usually) to search for ads in.
    :return: A tuple with the location as a tuple of x and y and the image that was searched in
    """
    pt = None
    try:
        loc, _ = find_image(image, TEMPLATE_AD_IMAGE)

        # Find the first (not the best) location
        pt = next(zip(*loc[::-1]), False)
        if pt:
            # Draw circle in the center of the match
            pt = (pt[0] + TEMPLATE_AD_IMAGE.shape[1] // 2, pt[1] + TEMPLATE_AD_IMAGE.shape[0] // 2)
            cv2.circle(image, pt, 16, (0, 0, 255, 255), 16)
    except Exception as e:
        print(e)

    # show_image(image)
    return pt, image


def handle_new_ad(ad: Tuple[Tuple[int, int], np.ndarray], device: AdbDevice, state: dict):
    """
    Handles a newly identified ad, whether or not there actually is one.
    This method also makes sure, that ads are closed again.
    Some ads open the Play Store, so this method also moves back to Nimses.

    :param ad: a tuple with the ad location (optional) and the screen shot.
    :param device: the Android device that should be controlled
    :param state: the global state of this application
    """

    def close_ad():
        """
        Closes the currently open ad (really just switches to the main activity)
        """
        if not state["ad_closed"] and state["ad_time"] < time():
            device.app_start(NIMSES_PACKAGE, NIMSES_MAIN_ACTIVITY)
            state["ad_closed"] = True
            logger.info("Closed ad")

    pt, image = ad
    app = device.current_app()
    logger.info("Current app %s", app)
    logger.info("Ad open? %s", not state["ad_closed"])

    if np.array_equal(image, state["last_image"]) \
            or state["last_ad"] > 10:
        output = device.shell(["am", "force-stop", NIMSES_PACKAGE])
        logger.info(f"[ADB] {output}")
        device.app_start(NIMSES_PACKAGE, NIMSES_MAIN_ACTIVITY)
        state["last_ad"] = 0
        state["ad_closed"] = True
        logger.info("Restarted Nimses")
    elif app["package"] == NIMSES_PACKAGE:
        if app["activity"] == NIMSES_MAIN_ACTIVITY:
            # Ad found, click it
            if pt:
                device.click(pt[0], pt[1])
                state["last_ad"] = 0
                state["ad_closed"] = True
                logger.info("Clicked %d, %d", pt[0], pt[1])
            # No ad, scroll along
            else:
                state["last_ad"] += 1
                w, h = device.window_size()
                sx = w // 2
                sy = h // 4 * 3
                dx = w // 2
                dy = h // 4 * 1
                device.swipe(sx, sy, dx, dy, 1.0)
        elif app["activity"] in NIMSES_AD_ACTIVITIES:
            if state["ad_closed"]:
                state["ad_time"] = time() + 35
                state["ad_closed"] = False
                logger.info("Close ad at: %s", strftime("%H:%M:%S", localtime(state["ad_time"])))
            else:
                close_ad()
    elif app["package"] == PLAYSTORE_PACKAGE:
        device.app_start(NIMSES_PACKAGE, NIMSES_MAIN_ACTIVITY)
        logger.info("Closed Play Store")
    else:
        device.app_start(NIMSES_PACKAGE, NIMSES_MAIN_ACTIVITY)


def main(device: AdbDevice = None):
    """
    Initializes the global state, sets up a ThreadPoolExecutor and launches the Nimses app.

    :param device: the device to control, if `None`, chooses the first and only device
    """
    executor = ThreadPoolExecutor(3)
    state = {
        "device": device if device is not None else adb.device(),  # The Android device to control
        "executor": executor,  # The executor where the magic happens
        "last_image": None,  # The last image
        "last_ad": 0,  # The amount of updates since the last ad
        "ad_closed": True,  # Whether the last ad has been closed
        "ad_time": 0  # When to close the app (in the future)
    }

    device.app_start(NIMSES_PACKAGE, NIMSES_MAIN_ACTIVITY)

    # Prepare future variable
    future = Future()
    future.set_result(None)
    while True:
        if future.done():
            future = executor.submit(
                lambda: log_error(
                    lambda: handle_new_ad(find_inline_ads(screen_shot_android(device)), device, state)))
        sleep(0.1)



import logging
import subprocess
from pathlib import Path
from typing import List

from adbutils import adb_path, adb

logger = logging.getLogger(__name__)


def execute_adb(args: List[str], **kwargs):
    cmdline = [adb_path()]
    cmdline.extend(args)
    cmdline = subprocess.list2cmdline(cmdline)
    return subprocess.check_output(cmdline,
                                   cwd=Path(adb_path()).parent,
                                   stderr=subprocess.STDOUT,
                                   encoding="utf-8",
                                   **kwargs)


def switch_to_tcpip(port: int = 5555):
    logger.info("Checking for connected devices")
    try:
        output = execute_adb(["wait-for-local-device"], timeout=1)
        logger.info("[ADB] %s", output)
        network_device = next(filter(lambda d: ":" in d.serial, adb.device_list()))
        if network_device is not None:
            logger.info("Found connected device: %s", network_device)
            return network_device
    except subprocess.TimeoutExpired:
        logger.warning("No networked device found.")
        pass

    logger.info("Switching to tcpip for USB connected device")
    logger.info("Waiting for device")
    output = execute_adb(["wait-for-usb-device"])
    logger.info("[ADB] %s", output)
    ip = adb.device().wlan_ip()
    logger.info(f"Trying to connect to {ip}:{port}")
    output = execute_adb(["-d", "tcpip", f"{port}"])
    logger.info("[ADB] %s", output)
    try:
        output = execute_adb(["disconnect", f"{ip}:{port}"])
    except subprocess.CalledProcessError as e:
        logger.warning("[ADB] %s", e)
    logger.info("[ADB] %s", output)
    output = adb.connect(f"{ip}:{port}")
    logger.info("[ADB] %s", output)
    output = execute_adb(["wait-for-local-device"])
    logger.info("[ADB] %s", output)
    logger.info("Please remove the USB device")
    return adb.device(f"{ip}:{port}")



from adbutils import AdbDevice


def send_raw(device: AdbDevice, command: str, buffer_size=1024) -> bytes:
    with device._client._connect() as c:
        c.send("host:transport:" + device.serial)
        c.check_okay()
        c.send(command)
        c.check_okay()
        buffer = b''
        while True:
            chunk = c.conn.recv(buffer_size)
            if not chunk:
                break
            buffer += chunk
        c.close()
        return buffer



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='nimses-adimate',
    version='1.2',
    packages=find_packages(),
    url='https://github.com/GiantTreeLP/nimses-adimate',
    license='MIT',
    author='GiantTree',
    author_email='gianttree@groundmc.net',
    description='This is a simple script that automatically clicks ads in the Nimses Android app.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3',
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Home Automation"
    ],
    install_requires=[
        'opencv-python',
        'numpy',
        'adbutils'
    ]
)



import logging
from typing import Callable

logger = logging.getLogger(__name__)


def log_error(function: Callable, *args):
    try:
        return function(*args)
    except Exception as e:
        logger.error(e)
